from typing import ClassVar

import os
import pytest

from typedllm import LLMModel, TypedPrompt, LLMSession, LLMRequest, Tool, create_tool_from_function
from typedtemplate import JinjaTemplateEngine, TypedTemplate
from pydantic import Field


@pytest.fixture(name="openai_key")
def get_openai_key() -> str:
    return os.getenv("OPENAI_API_KEY")


@pytest.fixture(name="model")
def fixture_model(openai_key: str) -> LLMModel:
    return LLMModel(
        name="gpt-4",
        api_key=openai_key,
    )


@pytest.fixture(name="prompt")
def get_prompt_value():
    return {
        "system": "You are an AI assistant. You must respond to questions and follow all instructions.",
        "prompt": "Please respond with just 'Hi' to this message"
    }


@pytest.fixture(name="hi_acompletion")
def hi_acompletion():
    return {"choices": [{"text": "Hi"}]}


@pytest.fixture(name="jinja_engine")
def get_template_engine():
    return JinjaTemplateEngine()


@pytest.fixture(name="llmprompt")
def get_prompt(jinja_engine: JinjaTemplateEngine):
    class PromptTemplate(TypedTemplate):
        template_engine = jinja_engine
        template_string = "What year was {{ city }} founded?"
        city: str = Field(description="The city to use in the prompt")

    return PromptTemplate


@pytest.fixture(name="instruction_template")
def get_instruction_template(jinja_engine: JinjaTemplateEngine):
    class InstructionTemplate(TypedTemplate):
        template_engine = jinja_engine
        template_string = "What year was {{ city }} founded?"
        city: str = Field(description="City name")
    return InstructionTemplate


@pytest.fixture(name="context_template")
def get_context_template(jinja_engine: JinjaTemplateEngine):
    class ContextTemplate(TypedTemplate):
        template_engine = jinja_engine
        template_string = "User Name: {{ username }}"
        username: str = Field(description="User's name")
    return ContextTemplate

@pytest.fixture(name="llmsession")
def get_session(model: LLMModel) -> LLMSession:
    return LLMSession(
        model=model,
        verbose=True,
    )


@pytest.fixture(name="llmrequest")
def get_request():
    return LLMRequest(force_text_response=True)


@pytest.fixture(name="llmtool")
def get_tool():
    class CityFoundingTool(Tool):
        """A tool to collect a city and its founding year."""
        city_name: str = Field(description="The name of the city")
        founded_year: int = Field(description="The year a city was founded")
    return CityFoundingTool


@pytest.fixture(name="FuncTool")
def get_function_tool():
    def TestTool(value: int) -> str:
        return str(value)
    return create_tool_from_function(TestTool)


@pytest.fixture(name="ClassTool")
def get_class_tool():
    class TestTool(Tool):
        value: int
    return TestTool



from pydantic import BaseModel
from typedllm.tool import Tool, create_tool_from_function


def test_basic_tool_creation(ClassTool):
    result = ClassTool.openapi_json()

    assert result == {
        "type": "function",
        "function": {
            "name": "TestTool",
            "description": "Tool called TestTool",
            "parameters": {
                "type": "object",
                "title": "TestTool",
                "description": "Tool called TestTool",
                "properties": {
                    "value": {
                        "title": "Value",
                        "type": "integer"
                    }
                },
                "required": ["value"]
            }
        }
    }


def test_func_versus_class_tool_creation(FuncTool, ClassTool):
    """
    This test should show that you can create a tool from a function or class and they are equivalent.
    """
    f = FuncTool.openapi_json()
    c = ClassTool.openapi_json()
    assert f == c


def test_tool_creation_from_function(FuncTool):
    result = FuncTool.openapi_json()

    assert FuncTool.__name__ == "TestTool"
    assert FuncTool.__doc__ == "Tool called TestTool"

    assert result == {
        "type": "function",
        "function": {
            "name": "TestTool",
            "description": "Tool called TestTool",
            "parameters": {
                "type": "object",
                "title": "TestTool",
                "description": "Tool called TestTool",
                "properties": {
                    "value": {
                        "title": "Value",
                        "type": "integer"
                    }
                },
                "required": ["value"]
            }
        }
    }



from typing import Callable

import typedllm
from typedllm.interop import litellm_request, async_litellm_request
from typedllm import LLMModel
import pytest


def acompletion_gen(response: dict) -> Callable:
    async def acompletion(*args, **kwargs):
        return response
    return acompletion


@pytest.mark.asyncio
async def test_basic_litellm_request(model: LLMModel, monkeypatch: pytest.MonkeyPatch, hi_acompletion: dict):
    monkeypatch.setattr(typedllm.interop, "acompletion", acompletion_gen(hi_acompletion))

    messages = [
        {"type": "message", "content": "Hi"}
    ]
    tools = []
    tool_choice = {}
    verbose = True

    response = await async_litellm_request(
        model.ssl_verify,
        model.name,
        model.max_retries,
        model.api_key,
        messages,
        tools,
        tool_choice,
        verbose,
        model.headers,
        model.organization,
        model.api_base
    )
    assert response == hi_acompletion, "Response should be the same as the fixture"


@pytest.mark.asyncio
async def test_no_msg_error(openai_key: str, model: LLMModel):
    messages = []
    tools = []
    tool_choice = {}
    verbose = True

    with pytest.raises(ValueError):
        await async_litellm_request(
            model.ssl_verify,
            model.name,
            model.max_retries,
            model.api_key,
            messages,
            tools,
            tool_choice,
            verbose,
            model.headers,
            model.organization,
            model.api_base
        )
        assert False, "Should not reach this point"



from typedtemplate import TypedTemplate
from typing import Type
from typedllm import TypedPrompt


def test_basic_prompt(instruction_template: Type[TypedTemplate], context_template: Type[TypedTemplate]):
    class TestPrompt(TypedPrompt):
        instructions: instruction_template
        context: context_template

    prompt = TestPrompt(
        instructions=instruction_template(city="San Francisco"),
        context=context_template(username="Johnny Test")
    )
    result = prompt.render()
    assert result is not None
    assert result == "# Instructions \nWhat year was San Francisco founded?\n# Context\nUser Name: Johnny Test"


def test_static_instructions(context_template: Type[TypedTemplate]):
    class TestPrompt(TypedPrompt):
        static_instructions = "What year was San Francisco founded?"
        context: context_template

    prompt = TestPrompt(
        context=context_template(username="Johnny Test")
    )
    result = prompt.render()
    assert result is not None
    assert result == "# Instructions \nWhat year was San Francisco founded?\n# Context\nUser Name: Johnny Test"



from typing import Type

import pytest

from pydantic import BaseModel, Field

from typedllm import (
    LLMModel,
    LLMSession,
    LLMRequest,
    LLMUserMessage,
    request,
    async_request,
    create_tool_from_function,
    Tool
)
from typedllm.client import typed_request, async_tool_from_json_str, sync_tool_from_json_str


@pytest.mark.asyncio
async def test_basic_request(openai_key: str):
    model = LLMModel(
        name="gpt-4-0125-preview",
        api_key=openai_key,
    )
    session = LLMSession(
        model=model,
    )
    request = LLMRequest(
        message=LLMUserMessage(
            content="What year was New York City founded?",
        ),
        force_text_response=True
    )
    session, response = await async_request(session, request)
    assert response.message.content.index("1624") > -1


class CityFoundingInfo(Tool):
    """This represents a city and includes the year the city was founded."""
    year: int = Field(description="The year the city was founded.")
    city: str = Field(description="The name of the city.")


def get_city_founding_history(city_info: CityFoundingInfo) -> str:
    return (f"Historical weather for {city_info.city} since its founding in {city_info.year}. "
            f"The weather is very nice. Average 80 degrees F.")


def test_make_tool_from_function():
    Tool = create_tool_from_function(get_city_founding_history)
    assert Tool.__name__ == "get_city_founding_history"
    assert "city_info" in Tool.model_fields
    assert Tool.model_fields["city_info"].annotation == CityFoundingInfo


@pytest.mark.asyncio
async def test_basic_tools_request(openai_key: str):
    model = LLMModel(
        name="gpt-4",
        api_key=openai_key,
    )
    tool: Type[Tool] = create_tool_from_function(get_city_founding_history)
    session = LLMSession(
        model=model,
        tools=[tool]
    )
    request = LLMRequest(
        message=LLMUserMessage(
            content="What is the founding story of New York City?",
        ),
        required_tool=tool,
    )
    session, response = await async_request(session, request)
    assert session
    assert response
    assert response.tool_calls[0].tool.__class__.__name__ == "get_city_founding_history"
    assert response.tool_calls[0].tool.city_info.city.index("New York") > -1
    assert response.tool_calls[0].tool.city_info.year == 1624


@pytest.mark.asyncio
async def test_typed_request(openai_key: str, llmprompt):
    class Year(Tool):
        year: int = Field(description="The year of the city's founding")

    model = LLMModel(
        name="gpt-4",
        api_key=openai_key,
    )
    tool = typed_request(model, llmprompt, Year, city="New York")
    assert tool
    assert tool.__class__.__name__ == "Year"
    assert tool.year == 1624


@pytest.mark.asyncio
async def test_async_tool_from_json_str(llmsession: LLMSession):
    broken_json_str = '{"year": 1624, "city": "New York"'
    result = await async_tool_from_json_str(llmsession, broken_json_str, CityFoundingInfo)
    assert result.year == 1624
    assert result.city == "New York"


@pytest.mark.asyncio
async def test_async_tool_from_json_str_missing_commas(llmsession: LLMSession):
    broken_json_str = '{"year": 1624 "city": "New York"}'
    result = await async_tool_from_json_str(llmsession, broken_json_str, CityFoundingInfo)
    assert result.year == 1624
    assert result.city == "New York"


def test_tool_from_json_str(llmsession: LLMSession):
    broken_json_str = '{"year": 1624, "city": "New York"'
    result = sync_tool_from_json_str(llmsession, broken_json_str, CityFoundingInfo)
    assert result.year == 1624
    assert result.city == "New York"


def test_tool_from_json_str_missing_commas(llmsession: LLMSession):
    broken_json_str = '{"year": 1624 "city": "New York"}'
    result = sync_tool_from_json_str(llmsession, broken_json_str, CityFoundingInfo)
    assert result.year == 1624
    assert result.city == "New York"
